/*	$NetBSD$	*/

/*
 * Copyright (c) 2000 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Kazuki Sakamoto <sakamoto@netbsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * YMF724/740/744/754 registers
 */

#ifndef _DEV_PCI_YMFREG_H_
#define	_DEV_PCI_YMFREG_H_

/*
 * PCI Config Registers
 */
#define	YMF_PCI_MBA		0x10
#define	YMF_PCI_LEGACY		0x40
#define	YMF_PCI_DSCTRL		0x48
# define YMF_DSCTRL_CRST	0x00000001
# define YMF_DSCTRL_WRST	0x00000004

/*
 * YMF registers
 */
#define	YMF_GLOBAL_CONTROL	0x0008
# define YMF_GLCTRL_HVE		0x0001
# define YMF_GLCTRL_HVIE	0x0002

#define	AC97_CMD_DATA		0x0060
#define	AC97_CMD_ADDR		0x0062
# define AC97_ID(id)		((id) << 8)
# define AC97_CMD_READ		0x8000
# define AC97_CMD_WRITE		0x0000
#define	AC97_STAT_DATA1		0x0064
#define	AC97_STAT_ADDR1		0x0066
#define	AC97_STAT_DATA2		0x0068
#define	AC97_STAT_ADDR2		0x006a
# define AC97_BUSY		0x8000

#define	YMF_DAC_OUT_VOLUME	0x0084
#define	YMF_DAC_OUT_VOL_L	0x0084
#define	YMF_DAC_OUT_VOL_R	0x0086

#define	YMF_ADC_SAMPLE_RATE	0x00c0
#define	YMF_REC_SAMPLE_RATE	0x00c4
#define	YMF_ADC_FORMAT		0x00c8
#define	YMF_REC_FORMAT		0x00cc
# define YMF_FORMAT_8BIT	0x01
# define YMF_FORMAT_STEREO	0x02

#define	YMF_STATUS		0x0100
# define YMF_STAT_ACT		0x00000001
# define YMF_STAT_WORK		0x00000002
# define YMF_STAT_TINT		0x00008000
# define YMF_STAT_INT		0x80000000
#define	YMF_CONTROL_SELECT	0x0104
# define YMF_CSEL		0x00000001
#define	YMF_MODE		0x0108
# define YMF_MODE_ACTV		0x00000001
# define YMF_MODE_ACTV2		0x00000002
# define YMF_MODE_TOUT		0x00008000
# define YMF_MODE_RESET		0x00010000
# define YMF_MODE_AC3		0x40000000
# define YMF_MODE_MUTE		0x80000000

#define	YMF_CONFIG		0x0114
# define YMF_DSP_DISABLE	0
# define YMF_DSP_SETUP		0x00000001

#define	YMF_PLAY_CTRLSIZE	0x0140
#define	YMF_REC_CTRLSIZE	0x0144
#define	YMF_EFFECT_CTRLSIZE	0x0148
#define	YMF_WORK_SIZE		0x014c
#define	YMF_MAPOF_REC		0x0150
# define YMF_RECSLOT_VALID	0x00000001
# define YMF_ADCSLOT_VALID	0x00000002
#define	YMF_MAPOF_EFFECT	0x0154
# define YMF_DL_VALID		0x00000001
# define YMF_DR_VALID		0x00000002
# define YMF_EFFECT1_VALID	0x00000004
# define YMF_EFFECT2_VALID	0x00000008
# define YMF_EFFECT3_VALID	0x00000010

#define	YMF_PLAY_CTRLBASE	0x0158
#define	YMF_REC_CTRLBASE	0x015c
#define	YMF_EFFECT_CTRLBASE	0x0160
#define	YMF_WORK_BASE		0x0164

#define	YMF_DSP_INSTRAM		0x1000
#define	YMF_CTRL_INSTRAM	0x4000

typedef enum {
	YMF_DS_1,
	YMF_DS_1E
} ymf_dstype_t;

#define	AC97_TIMEOUT		1000
#define	YMF_WORK_TIMEOUT	250000

#endif /* _DEV_PCI_YMFREG_H_ */
